/*
This file contains the initialization function for setting field and parameter values.

The only function here that should be called externally is initialize(). It sets field values in momentum space and then transforms them to position space. It also sets parameters such as the scale factor and its derivatives.
*/

#include "latticeeasy.h"

// Generate a uniform deviate between 0 and 1 using the Park-Miller minimum standard algorithm
#define randa 16807
#define randm 2147483647
#define randq 127773
#define randr 2836
float rand_uniform(void)
{
  static int next=seed;
  if(!(next>0)) // Guard against 0, negative, or other invalid seeds
  {
    printf("Invalid seed used in random number function. Using seed=1\n");
    next=1;
  }
  next = randa*(next%randq) - randr*(next/randq);
  if(next<0) next += randm;
  return ((float)next/(float)randm);
}
#undef randa
#undef randm
#undef randq
#undef randr

// Set the amplitude and phase of a mode of vacuum fluctuations
// Phase is set randomly (except for modes that must be real).
// Amplitude is set with a Rayleigh distribution about an rms value of 1/sqrt(omega) (times normalization terms).
void set_mode(float p2, float m2, float *field, float *deriv, int real)
{
  float phase,amplitude,rms_amplitude,omega;
  static float norm = rescale_A*rescale_B*sqrt(L*L*L/2)/(dx*dx*dx);
  static float hbterm = hubble_init*(rescale_r-1.);
  static int tachyonic=0; // Avoid printing the same error repeatedly

  phase=2.*pi*rand_uniform(); // Set phase randomly
  if(p2+m2>0.) // Check to avoid floating point errors
  {
    omega=sqrt(p2+m2); // Omega = Sqrt(p^2 + m^2)
    rms_amplitude=norm/sqrt(omega);
  }
  else
  {
    if(tachyonic==0)
      printf("Warning: Tachyonic mode(s) may be initialized inaccurately\n");
    omega=sqrt(p2); // If p^2 + m^2 < 0 use m^2=0
    rms_amplitude=(omega>0. ? norm/sqrt(omega) : 0.); // Avoid dividing by zero
    tachyonic=1;
  }
  amplitude=rms_amplitude*sqrt(log(1./rand_uniform())); // Amplitude = RMS amplitude x Rayleigh distributed random number
  field[0] = amplitude*cos(phase); // Re(field)
  field[1] = amplitude*sin(phase); // Im(field)
  deriv[0] = omega*field[1] + hbterm*field[0]; // Field derivative = (-i omega + (r-1) Hubble) Field
  deriv[1] = -omega*field[0] + hbterm*field[1];
  if(real==1) // For real modes set the imaginary parts to zero
  {
    field[1] = 0.;
    deriv[1] = 0.;
  }

  return;
}

/////////////////////////////////////////////////////
// Externally called function(s)
/////////////////////////////////////////////////////

// Set initial parameters and field values
void initialize()
{
  int fld,i,j,k,iconj,jconj; // iconj and jconj are the indices where conjugate modes to i and j are stored
  float px,py,pz,p2; // Components of momentum p_i and total squared momentum p2. px-pz are in units of grid spacing while p2 is converted to momentum units.
  float dp2=pw2(2.*pi/L); // Square of grid spacing in momentum space
  float mass_sq[nflds]; // Effective mass squared of fields
  float fnyquist[N][2*N],fdnyquist[N][2*N]; // Used by FFT routine to store modes with k=nyquist
  int arraysize[]={N,N,N}; // Array of grid size in all dimensions - used by FFT routine
  float initial_field_values[nflds]; // Initial value of fields (set to zero unless specified in parameters.h)
  float initial_field_derivs[nflds]; // Initial value of field derivatives (set to zero unless specified in parameters.h)
  float fsquared=0.,fdsquared=0.,ffd=0.,pot_energy=0.; // Sum(field^2), Sum(field_dot^2), Sum(f*f_dot), and potential energy - used for calculating the Hubble constant
  FILE *old_grid_; // Used for reading in previously generated data as initial conditions

  modelinitialize(); // This allows specific models to perform any needed initialization

  if((old_grid_=fopen("grid.img","rb"))) // If an old grid image can be opened use it to read in initial conditions
  {
    printf("Previously generated grid image found. Reading in data...\n");
    fread(&run_number,sizeof(run_number),1,old_grid_);
    fread(&t0,sizeof(t0),1,old_grid_);
    if(t0>=tf) // Check to make sure that the time in the old grid is earlier than the final time for this run
    {
      printf("A grid image file was found in this directory with values stored at t=%f. To continue that run set tf to a later time. To start a new run move or rename the file grid.img.\n",t0);
      exit(1);
    }
    fread(&a,sizeof(a),1,old_grid_);
    fread(&ad,sizeof(ad),1,old_grid_);
    fread(f,sizeof(float),nflds*gridsize,old_grid_);
    fread(fd,sizeof(float),nflds*gridsize,old_grid_);
    fclose(old_grid_);
    run_number++; // Iterate run_number to reflect the fact that this run is a continuation.
    printf("Data read. Resuming run at t=%f\n",t0);
    return;
  }

  // If no old grid image is found generate new initial conditions and set run_number=0
  printf("Generating initial conditions for new run at t=0\n");
  t0=0;
  run_number=0;

  // Check to make sure time step is small enough to satisfy Courant condition, dt/dx < 1/Sqrt(ndims)
  if(dt>dx/sqrt(3.))
  {
    printf("Time step too large. The ratio dt/dx is currently %f but for stability should never exceed 1/sqrt(3).\n",dt/dx);
    printf("Adjust dt to AT MOST %e, and preferably somewhat smaller than that.\n",dx/sqrt(3.));
    exit(1);
  }

  for(fld=0;fld<nflds;fld++)
  {
    // Set initial field values
    if(fld<(int)(sizeof initfield/sizeof(float))) // Use preset values for initial field values if given
      initial_field_values[fld]=initfield[fld];
    else // Otherwise initialize field to zero
      initial_field_values[fld]=0.;

    // Set initial field derivative values
    if(fld<(int)(sizeof initderivs/sizeof(float))) // Use preset values for initial field derivative values if given
      initial_field_derivs[fld]=initderivs[fld];
    else // Otherwise initialize derivatives to zero
      initial_field_derivs[fld]=0.;
  }

  // Set initial values of effective mass.
  effective_mass(mass_sq,initial_field_values);

  // Set initial value of Hubble constant - See the documentation for an explanation of the formulas used
  for(fld=0;fld<nflds;fld++) // Find sum of squares of fields and derivatives
  {
    fsquared += pw2(initial_field_values[fld]);
    fdsquared += pw2(initial_field_derivs[fld]);
    ffd += initial_field_values[fld]*initial_field_derivs[fld];
  }
  for(i=0;i<num_potential_terms;i++) // Find potential energy
    pot_energy += potential_energy(i,initial_field_values);
  hubble_init = sqrt( 3.*pw2(rescale_A)/(4.*pi)*fdsquared + 2.*pot_energy*(3.*pw2(rescale_A)/(4.*pi)-pw2(rescale_r)*fsquared) );
  hubble_init -= rescale_r*ffd;
  hubble_init /= 3.*pw2(rescale_A)/(4.*pi)-pw2(rescale_r)*fsquared;
  if(!(hubble_init>0.)) // Make sure Hubble isn't negative or undefined
  {
    printf("Error in calculating initial Hubble constant. Exiting.\n");
    exit(1);
  }
  if(expansion>0) // Set the derivative of the scale factor if the expansion of the universe is being used
    ad=hubble_init;

  for(fld=0;fld<nflds;fld++) // Set initial conditions for each field
  {
    // Loop over gridpoints.
    // px and py go over all mode values in wrap-around order, rising from 0 to N/2 and then from -N/2+1 to -1
    for(i=0;i<N;i++)
    {
      px=(i<=N/2 ? i : i-N); // x-component of momentum of modes at x=i
      iconj=(i==0 ? 0 : N-i); // Index where complex conjugates of modes at x=i are stored (only used for k=0 or k=N/2)
      for(j=0;j<N;j++)
      {
	py=(j<=N/2 ? j : j-N); // y-component of momentum of modes at y=j

	// Set all modes with 0<k<N/2. The complex conjugates of these modes do not need to be set.
	for(k=1;k<N/2;k++)
	{
	  pz=k; // z-component of momentum of modes at z=k
	  p2=dp2*(pw2(px)+pw2(py)+pw2(pz)); // Total momentum squared
	  set_mode(p2,mass_sq[fld],&f[fld][i][j][2*k],&fd[fld][i][j][2*k],0); // Set mode
	}

	// Set modes with k=0 or N/2.
	if(j>N/2 || (i>N/2 && (j==0 || j==N/2))) // The complex conjugates of these modes appear explicitly on the lattice and must be set to satisfy f(-p)=f(p)*
	{
	  jconj=(j==0 ? 0 : N-j); // Index where complex conjugates of modes at y=j are stored
	  // k=0
	  p2=dp2*(pw2(px)+pw2(py)); // Total momentum squared
	  set_mode(p2,mass_sq[fld],&f[fld][i][j][0],&fd[fld][i][j][0],0); // Set mode
	  f[fld][iconj][jconj][0]=f[fld][i][j][0]; // Set complex conjugate mode
	  f[fld][iconj][jconj][1]=-f[fld][i][j][1];
	  fd[fld][iconj][jconj][0]=fd[fld][i][j][0];
	  fd[fld][iconj][jconj][1]=-fd[fld][i][j][1];
	  // k=N/2
	  p2=dp2*(pw2(px)+pw2(py)+pw2(N/2)); // Total momentum squared
	  set_mode(p2,mass_sq[fld],&fnyquist[i][2*j],&fdnyquist[i][2*j],0); // Set mode
	  fnyquist[iconj][2*jconj]=fnyquist[i][2*j]; // Set complex conjugate mode
	  fnyquist[iconj][2*jconj+1]=-fnyquist[i][2*j+1];
	  fdnyquist[iconj][2*jconj]=fdnyquist[i][2*j];
	  fdnyquist[iconj][2*jconj+1]=-fdnyquist[i][2*j+1];
	}
	else if((i==0 || i==N/2) && (j==0 || j==N/2)) // The 8 "corners" of the lattice are set to real values
	{
	  p2=dp2*(pw2(px)+pw2(py)); // Total momentum squared for k=0
	  if(p2>0.) // Don't set the zeromode here (see below)
	    set_mode(p2,mass_sq[fld],&f[fld][i][j][0],&fd[fld][i][j][0],1); // The last argument specifies that a real value should be set
	  p2=dp2*(pw2(px)+pw2(py)+pw2(N/2)); // Total momentum squared for k=N/2
	  set_mode(p2,mass_sq[fld],&fnyquist[i][2*j],&fdnyquist[i][2*j],1); // The last argument specifies that a real value should be set
	}
      } // End of loop over j (y-index on lattice)
    } // End of loop over i (x-index on lattice)
    f[fld][0][0][0]=0.; // Set zeromode of field and derivative to zero (it gets added in position space)
    f[fld][0][0][1]=0.;
    fd[fld][0][0][0]=0.;
    fd[fld][0][0][1]=0.;

    fftrn((float *)f[fld],(float *)fnyquist,3,arraysize,-1); // Inverse Fourier transform of field
    fftrn((float *)fd[fld],(float *)fdnyquist,3,arraysize,-1); // Inverse Fourier transform of field derivatives

    // Add zeromode
    for(i=0;i<N;i++)
      for(j=0;j<N;j++)
	for(k=0;k<N;k++)
	{
	  f[fld][i][j][k] += initial_field_values[fld];
	  fd[fld][i][j][k] += initial_field_derivs[fld];
	}
  } // End loop over fields

  save(0); // Save field values and derived quantities
  output_parameters(); // Save information about the model and parameters and start the clock for timing the run

  printf("Finished initial conditions\n");
  
  return;
}







